/*
 * This file is part of XFellow.
 *
 * gfxdrv.c - SDL graphics driver
 * Copyright 2000-2001 by Dan Sutherland <dan@dillinja.org>
 *
 * TODO:
 * - Support for paletted (8-bit) modes
 * - Proper mode listing and selection
 * - Can't exit fullscreen mode
 */

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#include "portable.h"
#include "defs.h"
#include "fellow.h"
#include "draw.h"
#include "gfxdrvmutex.h"

static SDL_Surface *gfxDrvFramebuffer;
static Uint8 gfxDrvPalette[256][3];

SDL_mutex *gfxDrvMutex;

int gfxDrvRegisterDrawModes()
{
	draw_mode *mode;
	SDL_Rect **hwmodes;
	SDL_VideoInfo *vidinfo;
	int i;
	char ms[120];

	/* query available video modes */
	hwmodes = SDL_ListModes(NULL, SDL_FULLSCREEN);

	/* check for silly results */
	if(hwmodes == (SDL_Rect **)0) {
		fellowAddLog("gfxDrvRegisterDrawModes(): no hardware modes available\n");
		return 0;
	}
	if(hwmodes == (SDL_Rect **)-1) {
		fellowAddLog("gfxDrvRegisterDrawModes(): all modes available, don't know what to do\n");
		return 0;
	}

	if((mode = malloc(sizeof(draw_mode))) == NULL) {
		fellowAddLog("gfxDrvRegisterDrawModes(): FATAL: couldnt allocate memory for drawmode\n");
		return 0;
	}

	for(i = 0; hwmodes[i]; i++) {
		sprintf(ms, "gfxDrvRegisterDrawModes(): detected mode %dx%d\n", hwmodes[i]->w, hwmodes[i]->h);
		fellowAddLog(ms);
	}

	vidinfo = SDL_GetVideoInfo();
// FIXME: just add 800x600 for testing
//		mode->width = hwmodes[i]->w;
//		mode->height = hwmodes[i]->h;
		mode->width = 800;
		mode->height = 600;
		mode->bits = vidinfo->vfmt->BitsPerPixel;
		mode->windowed = TRUE;
		mode->redsize = 8 - vidinfo->vfmt->Rloss;
		mode->redpos = vidinfo->vfmt->Rshift;
		mode->greensize = 8 - vidinfo->vfmt->Gloss;
		mode->greenpos = vidinfo->vfmt->Gshift;
		mode->bluesize = 8 - vidinfo->vfmt->Bloss;
		mode->bluepos = vidinfo->vfmt->Bshift;
		mode->pitch = mode->width * vidinfo->vfmt->BytesPerPixel; /* Width of a line in bytes */
		drawModeAdd(mode);

	fellowAddLog("gfxDrvRegisterDrawModes(): added default video mode\n");
//	drawModesDump(); /* why the fsck does this give an undefined reference? it's right there in draw.c */

	return 1;
}

void gfxDrvSet8BitColor(ULO col, ULO r, ULO g, ULO b)
{
	/* we just store the colour in an array for later since the graphics surface
	   hasn't been allocated yet */
	gfxDrvPalette[col][0] = r;
	gfxDrvPalette[col][1] = g;
	gfxDrvPalette[col][2] = b;
}

void gfxDrvSetPalette()
{
}

void gfxDrvBufferFlip(void)
{
	SDL_Flip(gfxDrvFramebuffer);
}

void gfxDrvSetMode(draw_mode *dm, ULO vertical_scale)
{
	fellowAddLog("gfxDrvSetMode()\n");
}

UBY *gfxDrvValidateBufferPointer(void)
{
	/* the SDL docs say to make sure no syscalls should be made while a
		surface is locked, so we use a mutex to block the input
		thread (which uses some stdlib calls) */
	if(SDL_mutexP(gfxDrvMutex) == -1) {
		fellowAddLog("gfxDrvValidateBufferPointer(): FATAL: could not lock Mutex\n");
		return NULL;
	}
	SDL_LockSurface(gfxDrvFramebuffer);
	return gfxDrvFramebuffer->pixels;
}

void gfxDrvInvalidateBufferPointer(void)
{
	SDL_UnlockSurface(gfxDrvFramebuffer);
	if(SDL_mutexV(gfxDrvMutex) == -1) {
		fellowAddLog("gfxDrvInvalidateBufferPointer(): FATAL: could not unlock Mutex\n");
		fellowNastyExit();
	}
}

BOOLE gfxDrvEmulationStart(ULO maxbuffercount)
{
	BOOLE res;

	fellowAddLog("gfxDrvEmulationStart(): allocating video surface\n");

	if(res = (gfxDrvFramebuffer = SDL_SetVideoMode(800, 600, 16,
			SDL_SWSURFACE | SDL_DOUBLEBUF)) != NULL) {
//		fellowAddLog("gfxDrvEmulationStart(): video surface allocated OK\n");
		SDL_WM_SetCaption("XFellow", NULL);
	}

	fellowAddLog("gfxDrvEmulationStart(): leaving\n");
	return res;
}

ULO gfxDrvEmulationStartPost(void)
{
	return 1;
}

void gfxDrvEmulationStop(void)
{
	fellowAddLog("gfxDrvEmulationStop(): freeing video surface\n");

	SDL_FreeSurface(gfxDrvFramebuffer);
}

BOOLE gfxDrvStartup(void)
{
	BOOLE ret;

	fellowAddLog("gfxDrvStartup(): initialising graphics driver\n");

 	ret = SDL_InitSubSystem(SDL_INIT_VIDEO) == 0;

	if(ret) {
		ret = gfxDrvRegisterDrawModes();
	} else {
		fellowAddLog("gfxDrvStartup(): FATAL: SDL failed to initialise graphics subsystem\n");
	}

	/* create mutex */
	gfxDrvMutex = SDL_CreateMutex();

	return ret;
}

void gfxDrvShutdown(void)
{
	fellowAddLog("gfxDrvShutdown(): shutting down graphics driver\n");
	SDL_QuitSubSystem(SDL_INIT_VIDEO);

	SDL_DestroyMutex(gfxDrvMutex);
}

